@@ -19,17 +19,23 @@
 */
 
 #include <stdlib.h>
-#include <string.h>
 #include <time.h>
 
+#include "mpi/mpi.h"
+
 #include "mw_channel.h"
 #include "mw_cipher.h"
 #include "mw_debug.h"
 #include "mw_session.h"
 
 
+struct mwMpi {
+  mw_mp_int i;
+};
+
+
 /** From RFC2268 */
-static unsigned char PT[] = {
+static guchar PT[] = {
   0xD9, 0x78, 0xF9, 0xC4, 0x19, 0xDD, 0xB5, 0xED,
   0x28, 0xE9, 0xFD, 0x79, 0x4A, 0xA0, 0xD8, 0x9D,
   0xC6, 0x7E, 0x37, 0x83, 0x2B, 0x76, 0x53, 0x8E,
@@ -65,42 +71,187 @@ static unsigned char PT[] = {
 };
 
 
-void rand_key(char *key, gsize keylen) {
-  srand(clock());
+/** prime number used in DH exchange */
+static guchar dh_prime[] = {
+  0xCF, 0x84, 0xAF, 0xCE, 0x86, 0xDD, 0xFA, 0x52,
+  0x7F, 0x13, 0x6D, 0x10, 0x35, 0x75, 0x28, 0xEE,
+  0xFB, 0xA0, 0xAF, 0xEF, 0x80, 0x8F, 0x29, 0x17,
+  0x4E, 0x3B, 0x6A, 0x9E, 0x97, 0x00, 0x01, 0x71,
+  0x7C, 0x8F, 0x10, 0x6C, 0x41, 0xC1, 0x61, 0xA6,
+  0xCE, 0x91, 0x05, 0x7B, 0x34, 0xDA, 0x62, 0xCB,
+  0xB8, 0x7B, 0xFD, 0xC1, 0xB3, 0x5C, 0x1B, 0x91,
+  0x0F, 0xEA, 0x72, 0x24, 0x9D, 0x56, 0x6B, 0x9F
+};
+
+
+/** base used in DH exchange */
+#define DH_BASE  3
+
+
+struct mwMpi *mwMpi_new() {
+  struct mwMpi *i;
+  i = g_new0(struct mwMpi, 1);
+  mw_mp_init(&i->i);
+  return i;
+}
+
+
+void mwMpi_free(struct mwMpi *i) {
+  if(! i) return;
+  mw_mp_clear(&i->i);
+  g_free(i);
+}
+
+
+static void mwInitDHPrime(mw_mp_int *i) {
+  mw_mp_init(i);
+  mw_mp_read_unsigned_bin(i, dh_prime, 64);
+}
+
+
+void mwMpi_setDHPrime(struct mwMpi *i) {
+  g_return_if_fail(i != NULL);
+  mw_mp_read_unsigned_bin(&i->i, dh_prime, 64);
+}
+
+
+static void mwInitDHBase(mw_mp_int *i) {
+  mw_mp_init(i);
+  mw_mp_set_int(i, DH_BASE);
+}
+
+
+void mwMpi_setDHBase(struct mwMpi *i) {
+  g_return_if_fail(i != NULL);
+  mw_mp_set_int(&i->i, DH_BASE);
+}
+
+
+static void mw_mp_set_rand(mw_mp_int *i, guint bits) {
+  size_t len, l;
+  guchar *buf;
+
+  l = len = (bits / 8) + 1;
+  buf = g_malloc(len);
+
+  srand(time(NULL));
+  while(l--) buf[l] = rand() & 0xff;
+
+  buf[0] &= (0xff >> (8 - (bits % 8)));
+
+  mw_mp_read_unsigned_bin(i, buf, len);
+  g_free(buf);
+}
+
+
+static void mwDHRandKeypair(mw_mp_int *private_key, mw_mp_int *public_key) {
+  mw_mp_int prime, base;
+ 
+  mwInitDHPrime(&prime);
+  mwInitDHBase(&base);
+
+  mw_mp_set_rand(private_key, 512);
+  mw_mp_exptmod(&base, private_key, &prime, public_key);
+
+  mw_mp_clear(&prime);
+  mw_mp_clear(&base);
+}
+
+
+void mwMpi_randDHKeypair(struct mwMpi *private_key, struct mwMpi *public_key) {
+  g_return_if_fail(private_key != NULL);
+  g_return_if_fail(public_key != NULL);
+
+  mwDHRandKeypair(&private_key->i, &public_key->i);
+}
+
+
+static void mwDHCalculateShared(mw_mp_int *shared_key, mw_mp_int *remote_key,
+				mw_mp_int *private_key) {
+  mw_mp_int prime;
+ 
+  mwInitDHPrime(&prime);
+  mw_mp_exptmod(remote_key, private_key, &prime, shared_key);
+  mw_mp_clear(&prime);
+}
+
+
+void mwMpi_calculateDHShared(struct mwMpi *shared_key, struct mwMpi *remote_key,
+			     struct mwMpi *private_key) {
+
+  g_return_if_fail(shared_key != NULL);
+  g_return_if_fail(remote_key != NULL);
+  g_return_if_fail(private_key != NULL);
+
+  mwDHCalculateShared(&shared_key->i, &remote_key->i, &private_key->i);
+}
+
+
+static void mwDHImportKey(mw_mp_int *key, struct mwOpaque *o) {
+  mw_mp_read_unsigned_bin(key, o->data, o->len);
+}
+
+
+void mwMpi_import(struct mwMpi *i, struct mwOpaque *o) {
+  g_return_if_fail(i != NULL);
+  g_return_if_fail(o != NULL);
+
+  mwDHImportKey(&i->i, o);
+}
+
+
+static void mwDHExportKey(mw_mp_int *key, struct mwOpaque *o) {
+  o->len = mw_mp_unsigned_bin_size(key);
+  o->data = g_malloc0(o->len);
+  mw_mp_to_unsigned_bin(key, o->data);
+}
+
+
+void mwMpi_export(struct mwMpi *i, struct mwOpaque *o) {
+  g_return_if_fail(i != NULL);
+  g_return_if_fail(o != NULL);
+  
+  mwDHExportKey(&i->i, o);
+}
+
+
+void mwKeyRandom(guchar *key, gsize keylen) {
+  g_return_if_fail(key != NULL);
+
+  srand(time(NULL));
   while(keylen--) key[keylen] = rand() & 0xff;
 }
 
 
-void mwIV_init(char *iv) {
-  static unsigned char normal_iv[] = {
-    0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef
-  };
-  memcpy(iv, normal_iv, 8);
-
-  /*
-  *iv++ = 0x01; *iv++ = 0x23;
-  *iv++ = 0x45; *iv++ = 0x67;
-  *iv++ = 0x89; *iv++ = 0xab;
-  *iv++ = 0xcd; *iv   = 0xef;
-  */
+void mwIV_init(guchar *iv) {
+  iv[0] = 0x01;
+  iv[1] = 0x23;
+  iv[2] = 0x45;
+  iv[3] = 0x67;
+  iv[4] = 0x89;
+  iv[5] = 0xab;
+  iv[6] = 0xcd;
+  iv[7] = 0xef;
 }
 
 
 /* This does not seem to produce the same results as normal RC2 key
    expansion would, but it works, so eh. It might be smart to farm
    this out to mozilla or openssl */
-void mwKeyExpand(int *ekey, const char *key, gsize keylen) {
-  char tmp[128];
+void mwKeyExpand(int *ekey, const guchar *key, gsize keylen) {
+  guchar tmp[128];
   int i, j;
 
-  /*
-  g_message("expanding key from:");
-  pretty_print(key, keylen);
-  */
+  g_return_if_fail(keylen > 0);
+  g_return_if_fail(key != NULL);
 
   if(keylen > 128) keylen = 128;
-  memcpy(tmp, key, keylen);
 
+  /* fill the first chunk with what key bytes we have */
+  for(i = keylen; i--; tmp[i] = key[i]);
+  /* memcpy(tmp, key, keylen); */
+
+  /* build the remaining key from the given data */
   for(i = 0; keylen < 128; i++) {
     tmp[keylen] = PT[ (tmp[keylen - 1] + tmp[i]) & 0xff ];
     keylen++;
@@ -116,14 +267,16 @@ void mwKeyExpand(int *ekey, const char *key, gsize keylen) {
 
 
 /* normal RC2 encryption given a full 128-byte (as 64 ints) key */
-static void mwEncryptBlock(const int *ekey, char *out) {
-  int a = (out[7] << 8) | (out[6] & 0xff);
-  int b = (out[5] << 8) | (out[4] & 0xff);
-  int c = (out[3] << 8) | (out[2] & 0xff);
-  int d = (out[1] << 8) | (out[0] & 0xff);
+static void mwEncryptBlock(const int *ekey, guchar *out) {
 
+  int a, b, c, d;
   int i, j;
 
+  a = (out[7] << 8) | (out[6] & 0xff);
+  b = (out[5] << 8) | (out[4] & 0xff);
+  c = (out[3] << 8) | (out[2] & 0xff);
+  d = (out[1] << 8) | (out[0] & 0xff);
+
   for(i = 0; i < 16; i++) {
     j = i * 4;
 
@@ -158,20 +311,21 @@ static void mwEncryptBlock(const int *ekey, char *out) {
 }
 
 
-void mwEncryptExpanded(const int *ekey, char *iv,
+void mwEncryptExpanded(const int *ekey, guchar *iv,
 		       struct mwOpaque *in_data,
 		       struct mwOpaque *out_data) {
 
-  char *i = in_data->data;
+  guchar *i = in_data->data;
   gsize i_len = in_data->len;
 
-  char *o;
+  guchar *o;
   gsize o_len;
 
   int x, y;
 
   /* pad upwards to a multiple of 8 */
-  o_len = (i_len & -8) + 8;
+  /* o_len = (i_len & -8) + 8; */
+  o_len = i_len + (8 - (i_len % 8));
   o = g_malloc(o_len);
 
   out_data->data = o;
@@ -181,20 +335,23 @@ void mwEncryptExpanded(const int *ekey, char *iv,
   y = o_len - i_len;
 
   /* copy in to out, and write padding bytes */
-  memcpy(o, i, i_len);
-  memset(o + i_len, y, y);
+  for(x = i_len; x--; o[x] = i[x]);
+  for(x = i_len; x < o_len; o[x++] = y);
+  /* memcpy(o, i, i_len);
+     memset(o + i_len, y, y); */
 
   /* encrypt in blocks */
   for(x = o_len; x > 0; x -= 8) {
     for(y = 8; y--; o[y] ^= iv[y]);
     mwEncryptBlock(ekey, o);
-    memcpy(iv, o, 8);
+    for(y = 8; y--; iv[y] = o[y]);
+    /* memcpy(iv, o, 8); */
     o += 8;
   }
 }
 
 
-void mwEncrypt(const char *key, gsize keylen, char *iv,
+void mwEncrypt(const guchar *key, gsize keylen, guchar *iv,
 	       struct mwOpaque *in, struct mwOpaque *out) {
 
   int ekey[64];
@@ -203,14 +360,16 @@ void mwEncrypt(const char *key, gsize keylen, char *iv,
 }
 
 
-static void mwDecryptBlock(const int *ekey, char *out) {
-  int a = (out[7] << 8) | (out[6] & 0xff);
-  int b = (out[5] << 8) | (out[4] & 0xff);
-  int c = (out[3] << 8) | (out[2] & 0xff);
-  int d = (out[1] << 8) | (out[0] & 0xff);
+static void mwDecryptBlock(const int *ekey, guchar *out) {
 
+  int a, b, c, d;
   int i, j;
 
+  a = (out[7] << 8) | (out[6] & 0xff);
+  b = (out[5] << 8) | (out[4] & 0xff);
+  c = (out[3] << 8) | (out[2] & 0xff);
+  d = (out[1] << 8) | (out[0] & 0xff);
+
   for(i = 16; i--; ) {
     j = i * 4 + 3;
 
@@ -245,24 +404,29 @@ static void mwDecryptBlock(const int *ekey, char *out) {
 }
 
 
-void mwDecryptExpanded(const int *ekey, char *iv,
+void mwDecryptExpanded(const int *ekey, guchar *iv,
 		       struct mwOpaque *in_data,
 		       struct mwOpaque *out_data) {
 
-  char *i = in_data->data;
+  guchar *i = in_data->data;
   gsize i_len = in_data->len;
 
-  char *o;
+  guchar *o;
   gsize o_len;
 
   int x, y;
 
-  /* this doesn't check to ensure that in_data->len is a multiple of
-     8, which is damn well ought to be. */
+  if(i_len % 8) {
+    /* this doesn't check to ensure that in_data->len is a multiple of
+       8, which is damn well ought to be. */
+    g_warning("attempting decryption of mis-sized data, %u bytes",
+	      (guint) i_len);
+  }
 
   o = g_malloc(i_len);
   o_len = i_len;
-  memcpy(o, i, i_len);
+  for(x = i_len; x--; o[x] = i[x]);
+  /* memcpy(o, i, i_len); */
 
   out_data->data = o;
   out_data->len = o_len;
@@ -273,7 +437,8 @@ void mwDecryptExpanded(const int *ekey, char *iv,
 
     /* modify the initialization vector */
     for(y = 8; y--; o[y] ^= iv[y]);
-    memcpy(iv, i, 8);
+    for(y = 8; y--; iv[y] = i[y]);
+    /* memcpy(iv, i, 8); */
     i += 8;
     o += 8;
   }
@@ -284,7 +449,7 @@ void mwDecryptExpanded(const int *ekey, char *iv,
 }
 
 
-void mwDecrypt(const char *key, gsize keylen, char *iv,
+void mwDecrypt(const guchar *key, gsize keylen, guchar *iv,
 	       struct mwOpaque *in, struct mwOpaque *out) {
 
   int ekey[64];
@@ -293,6 +458,7 @@ void mwDecrypt(const char *key, gsize keylen, char *iv,
 }
 
 
+
 struct mwCipher_RC2_40 {
   struct mwCipher cipher;
   int session_key[64];
@@ -303,8 +469,8 @@ struct mwCipher_RC2_40 {
 struct mwCipherInstance_RC2_40 {
   struct mwCipherInstance instance;
   int incoming_key[64];
-  char outgoing_iv[8];
-  char incoming_iv[8];
+  guchar outgoing_iv[8];
+  guchar incoming_iv[8];
 };
 
 
@@ -358,27 +524,29 @@ static int decrypt_RC2_40(struct mwCipherInstance *ci,
 }
 
 
-static struct mwCipherInstance *new_instance_RC2_40(struct mwCipher *cipher,
-						    struct mwChannel *chan) {
+static struct mwCipherInstance *
+new_instance_RC2_40(struct mwCipher *cipher,
+		    struct mwChannel *chan) {
+
   struct mwCipher_RC2_40 *cr;
   struct mwCipherInstance_RC2_40 *cir;
   struct mwCipherInstance *ci;
 
   cr = (struct mwCipher_RC2_40 *) cipher;
 
+  /* a bit of lazy initialization here */
+  if(! cr->ready) {
+    struct mwLoginInfo *info = mwSession_getLoginInfo(cipher->session);
+    mwKeyExpand(cr->session_key, (guchar *) info->login_id, 5);
+    cr->ready = TRUE;
+  }
+
   cir = g_new0(struct mwCipherInstance_RC2_40, 1);
   ci = &cir->instance;
 
   ci->cipher = cipher;
   ci->channel = chan;
 
-  /* a bit of lazy initialization here */
-  if(! cr->ready) {
-    struct mwLoginInfo *info = mwSession_getLoginInfo(cipher->session);
-    mwKeyExpand(cr->session_key, info->login_id, 5);
-    cr->ready = TRUE;
-  }
-
   mwIV_init(cir->incoming_iv);
   mwIV_init(cir->outgoing_iv);
 
@@ -387,24 +555,40 @@ static struct mwCipherInstance *new_instance_RC2_40(struct mwCipher *cipher,
 
 
 static struct mwEncryptItem *new_item_RC2_40(struct mwCipherInstance *ci) {
-  struct mwEncryptItem *e = g_new0(struct mwEncryptItem, 1);
+  struct mwEncryptItem *e;
+
+  e = g_new0(struct mwEncryptItem, 1);
   e->id = mwCipher_RC2_40;
   return e;
 }
 
 
-static void accept_RC2_40(struct mwCipherInstance *ci) {
-  struct mwCipherInstance_RC2_40 *cir;
-  struct mwLoginInfo *info = mwChannel_getUser(ci->channel);
-
-  cir = (struct mwCipherInstance_RC2_40 *) ci;
-  mwKeyExpand(cir->incoming_key, info->login_id, 5);
+static struct mwEncryptItem *
+offer_RC2_40(struct mwCipherInstance *ci) {
+  return new_item_RC2_40(ci);
 }
 
 
 static void accepted_RC2_40(struct mwCipherInstance *ci,
 			    struct mwEncryptItem *item) {
-  accept_RC2_40(ci);
+
+  struct mwCipherInstance_RC2_40 *cir;
+  struct mwLoginInfo *info;
+
+  cir = (struct mwCipherInstance_RC2_40 *) ci;
+  info = mwChannel_getUser(ci->channel);
+
+  if(info->login_id) {
+    mwKeyExpand(cir->incoming_key, (guchar *) info->login_id, 5);
+  }
+}
+
+
+static struct mwEncryptItem *
+accept_RC2_40(struct mwCipherInstance *ci) {
+
+  accepted_RC2_40(ci, NULL);
+  return new_item_RC2_40(ci);
 }
 
 
@@ -417,7 +601,8 @@ struct mwCipher *mwCipher_new_RC2_40(struct mwSession *s) {
   c->get_name = get_name_RC2_40;
   c->get_desc = get_desc_RC2_40;
   c->new_instance = new_instance_RC2_40;
-  c->new_item = new_item_RC2_40;
+
+  c->offer = offer_RC2_40;
 
   c->accepted = accepted_RC2_40;
   c->accept = accept_RC2_40;
@@ -429,6 +614,202 @@ struct mwCipher *mwCipher_new_RC2_40(struct mwSession *s) {
 }
 
 
+struct mwCipher_RC2_128 {
+  struct mwCipher cipher;
+  mw_mp_int private_key;
+  struct mwOpaque public_key;
+};
+
+
+struct mwCipherInstance_RC2_128 {
+  struct mwCipherInstance instance;
+  int shared[64];      /* shared secret determined via DH exchange */
+  guchar outgoing_iv[8];
+  guchar incoming_iv[8];
+};
+
+
+static const char *get_name_RC2_128() {
+  return "RC2/128 Cipher";
+}
+
+
+static const char *get_desc_RC2_128() {
+  return "RC2, DH shared secret key";
+}
+
+
+static struct mwCipherInstance *
+new_instance_RC2_128(struct mwCipher *cipher,
+		     struct mwChannel *chan) {
+
+  struct mwCipher_RC2_128 *cr;
+  struct mwCipherInstance_RC2_128 *cir;
+  struct mwCipherInstance *ci;
+
+  cr = (struct mwCipher_RC2_128 *) cipher;
+
+  cir = g_new0(struct mwCipherInstance_RC2_128, 1);
+  ci = &cir->instance;
+  
+  ci->cipher = cipher;
+  ci->channel = chan;
+
+  mwIV_init(cir->incoming_iv);
+  mwIV_init(cir->outgoing_iv);
+
+  return ci;
+}
+
+
+static void offered_RC2_128(struct mwCipherInstance *ci,
+			    struct mwEncryptItem *item) {
+  
+  mw_mp_int remote_key;
+  mw_mp_int shared;
+  struct mwOpaque sho = { 0, 0 };
+
+  struct mwCipher *c;
+  struct mwCipher_RC2_128 *cr;
+  struct mwCipherInstance_RC2_128 *cir;
+
+  c = ci->cipher;
+  cr = (struct mwCipher_RC2_128 *) c;
+  cir = (struct mwCipherInstance_RC2_128 *) ci;
+
+  mw_mp_init(&remote_key);
+  mw_mp_init(&shared);
+
+  mwDHImportKey(&remote_key, &item->info);
+  mwDHCalculateShared(&shared, &remote_key, &cr->private_key);
+  mwDHExportKey(&shared, &sho);
+
+  /* key expanded from the last 16 bytes of the DH shared secret. This
+     took me forever to figure out. 16 bytes is 128 bit. */
+  /* the sh_len-16 is important, because the key len could
+     hypothetically start with 8bits or more unset, meaning the
+     exported key might be less than 64 bytes in length */
+  mwKeyExpand(cir->shared, sho.data+(sho.len-16), 16);
+  
+  mw_mp_clear(&remote_key);
+  mw_mp_clear(&shared);
+  mwOpaque_clear(&sho);
+}
+
+
+static struct mwEncryptItem *
+offer_RC2_128(struct mwCipherInstance *ci) {
+
+  struct mwCipher *c;
+  struct mwCipher_RC2_128 *cr;
+  struct mwEncryptItem *ei;
+
+  c = ci->cipher;
+  cr = (struct mwCipher_RC2_128 *) c;
+
+  ei = g_new0(struct mwEncryptItem, 1);
+  ei->id = mwCipher_RC2_128;
+  mwOpaque_clone(&ei->info, &cr->public_key);
+
+  return ei;
+}			  
+
+
+static void accepted_RC2_128(struct mwCipherInstance *ci,
+			     struct mwEncryptItem *item) {
+
+  return offered_RC2_128(ci, item);
+}
+
+
+static struct mwEncryptItem *
+accept_RC2_128(struct mwCipherInstance *ci) {
+
+  return offer_RC2_128(ci);
+}
+
+
+static int encrypt_RC2_128(struct mwCipherInstance *ci,
+			   struct mwOpaque *data) {
+
+  struct mwCipherInstance_RC2_128 *cir;
+  struct mwOpaque o = { 0, 0 };
+
+  cir = (struct mwCipherInstance_RC2_128 *) ci;
+
+  mwEncryptExpanded(cir->shared, cir->outgoing_iv, data, &o);
+
+  mwOpaque_clear(data);
+  data->data = o.data;
+  data->len = o.len;
+
+  return 0;
+}
+
+
+static int decrypt_RC2_128(struct mwCipherInstance *ci,
+			   struct mwOpaque *data) {
+
+  struct mwCipherInstance_RC2_128 *cir;
+  struct mwOpaque o = { 0, 0 };
+
+  cir = (struct mwCipherInstance_RC2_128 *) ci;
+
+  mwDecryptExpanded(cir->shared, cir->incoming_iv, data, &o);
+
+  mwOpaque_clear(data);
+  data->data = o.data;
+  data->len = o.len;
+
+  return 0;
+}
+
+
+static void clear_RC2_128(struct mwCipher *c) {
+  struct mwCipher_RC2_128 *cr;
+  cr = (struct mwCipher_RC2_128 *) c;
+
+  mw_mp_clear(&cr->private_key);
+  mwOpaque_clear(&cr->public_key);
+}
+
+
+struct mwCipher *mwCipher_new_RC2_128(struct mwSession *s) {
+  struct mwCipher_RC2_128 *cr;
+  struct mwCipher *c;
+
+  mw_mp_int pubkey;
+
+  cr = g_new0(struct mwCipher_RC2_128, 1);
+  c = &cr->cipher;
+
+  c->session = s;
+  c->type = mwCipher_RC2_128;
+  c->get_name = get_name_RC2_128;
+  c->get_desc = get_desc_RC2_128;
+  c->new_instance = new_instance_RC2_128;
+
+  c->offered = offered_RC2_128;
+  c->offer = offer_RC2_128;
+
+  c->accepted = accepted_RC2_128;
+  c->accept = accept_RC2_128;
+
+  c->encrypt = encrypt_RC2_128;
+  c->decrypt = decrypt_RC2_128;
+
+  c->clear = clear_RC2_128;
+  
+  mw_mp_init(&cr->private_key);
+  mw_mp_init(&pubkey);
+  mwDHRandKeypair(&cr->private_key, &pubkey);
+  mwDHExportKey(&pubkey, &cr->public_key);
+  mw_mp_clear(&pubkey);
+
+  return c;
+}
+
+
 struct mwSession *mwCipher_getSession(struct mwCipher *cipher) {
   g_return_val_if_fail(cipher != NULL, NULL);
   return cipher->session;
@@ -437,8 +818,8 @@ struct mwSession *mwCipher_getSession(struct mwCipher *cipher) {
 
 guint16 mwCipher_getType(struct mwCipher *cipher) {
   /* oh man, this is a bad failover... who the hell decided to make
-     zero a real cipher id?? */
-  g_return_val_if_fail(cipher != NULL, 0x00);
+     zero a real cipher id? */
+  g_return_val_if_fail(cipher != NULL, 0xffff);
   return cipher->type;
 }
 
@@ -488,19 +869,6 @@ struct mwChannel *mwCipherInstance_getChannel(struct mwCipherInstance *ci) {
 }
 
 
-struct mwEncryptItem *mwCipherInstance_newItem(struct mwCipherInstance *ci) {
-  struct mwCipher *cipher;
-
-  g_return_val_if_fail(ci != NULL, NULL);
-  cipher = ci->cipher;
-
-  g_return_val_if_fail(cipher != NULL, NULL);
-  g_return_val_if_fail(cipher->new_item != NULL, NULL);
-
-  return cipher->new_item(ci);
-}
-
-
 void mwCipherInstance_offered(struct mwCipherInstance *ci,
 			      struct mwEncryptItem *item) {
   struct mwCipher *cipher;
@@ -514,15 +882,16 @@ void mwCipherInstance_offered(struct mwCipherInstance *ci,
 }
 
 
-void mwCipherInstance_offer(struct mwCipherInstance *ci) {
+struct mwEncryptItem *
+mwCipherInstance_offer(struct mwCipherInstance *ci) {
   struct mwCipher *cipher;
 
-  g_return_if_fail(ci != NULL);
+  g_return_val_if_fail(ci != NULL, NULL);
 
   cipher = ci->cipher;
-  g_return_if_fail(cipher != NULL);
+  g_return_val_if_fail(cipher != NULL, NULL);
 
-  if(cipher->offer) cipher->offer(ci);
+  return cipher->offer(ci);
 }
 
 
@@ -539,15 +908,16 @@ void mwCipherInstance_accepted(struct mwCipherInstance *ci,
 }
 
 
-void mwCipherInstance_accept(struct mwCipherInstance *ci) {
+struct mwEncryptItem *
+mwCipherInstance_accept(struct mwCipherInstance *ci) {
   struct mwCipher *cipher;
 
-  g_return_if_fail(ci != NULL);
+  g_return_val_if_fail(ci != NULL, NULL);
 
   cipher = ci->cipher;
-  g_return_if_fail(cipher != NULL);
+  g_return_val_if_fail(cipher != NULL, NULL);
 
-  if(cipher->accept) cipher->accept(ci);
+  return cipher->accept(ci);
 }
 
 
